#include "NebulaDevice.h"
#include "NebulaP2PTunnelAPIs.h"
#include "P2PTunnelAgent.h"
#include "disposable_params.h"
#include <ctype.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>

#define ENABLE_TUNNEL_SERVICE 0

#define MAX_SERVER_CONNECT_NUM 4
#define NEBULA_BIND_TIMEOUT_MS 10000
#define IOTC_CONNECT_TIMEOUT_MS 10000
#define TUNNEL_CONNECT_TIMEOUT_MS 10000

#define HTTP_MAPPING_LOCAL_PORT 10000
#define HTTP_MAPPING_REMOTE_PORT 80

#define HTTPS_MAPPING_LOCAL_PORT 15000

#define SSH_MAPPING_LOCAL_PORT 20000
#define SSH_MAPPING_REMOTE_PORT 22

#define TELNET_MAPPING_LOCAL_PORT 30000
#define TELNET_MAPPING_REMOTE_PORT 23

typedef struct st_NebulaTunnelInfo {
    NebulaClientCtx *pClientCtx;
    bool bAgentSuccess;
    int nHttpIndex;
    int nHttpsIndex;
    int nTelnetIndex;
    int nSshIndex;
    int nRetryConnectFailedCnt;
    int nSID;
    bool bException;
} NebulaTunnelInfo;

//########################################################
//# Nebula Client
//########################################################
static unsigned int gNebulaBindAbort = 0;

static NebulaClientCtx *nebulaclientCreate(const char *udid, const char *pin_code) {
    int ret = 0;
    NebulaClientCtx *client_ctx = NULL;
    TUTKJsonObject *json_args = NULL;

    ret = Nebula_Initialize();
    printf("Nebula_Initialize() ret[%d]\n", ret);
    if (ret != NEBULA_ER_NoERROR) {
        return NULL;
    }

    ret = Nebula_Client_New(udid, &client_ctx);
    printf("Nebula_Client_New() ret[%d]\n", ret);
    if (ret != NEBULA_ER_NoERROR) {
        Nebula_DeInitialize();
        return NULL;
    }

    ret = Nebula_Client_Bind(client_ctx, pin_code, &json_args, NEBULA_BIND_TIMEOUT_MS, &gNebulaBindAbort);
    printf("Nebula_Client_Bind() ret[%d]\n", ret);
    if (ret != NEBULA_ER_NoERROR) {
        Nebula_Client_Delete(client_ctx);
        Nebula_DeInitialize();
        return NULL;
    }

    printf("[Bind_Response]\n%s\n", TUTK_Json_Obj_To_String(json_args));

    ret = Nebula_Client_Free_Bind_Response(json_args);
    if (ret != NEBULA_ER_NoERROR) {
        printf("Nebula_Client_Free_Bind_Response() ret[%d]\n", ret);
    }

    return client_ctx;
}

//########################################################
//# IOTC Client
//########################################################
static char gRealm[MAX_REALM_LENGTH + 1] = "hausetopia-us";
static unsigned int gIOTCConnectAbort = 0;

static int iotcClientCreateByNebula(NebulaClientCtx *client_ctx, const char *am_token, NebulaTunnelInfo *info, unsigned int *abort) {
    int ret = 0;
    
    ret = IOTC_Client_Connect_By_Nebula(client_ctx, am_token, gRealm, IOTC_CONNECT_TIMEOUT_MS, abort);
    printf("IOTC_Client_Connect_By_Nebula() ret[%d]\n", ret);
    if (ret < 0) {
        return -1;
    }

    info->nSID = ret;
    return 0;
}

//########################################################
//# P2PTunnel Agent
//########################################################

static bool tunnelExceptionCheck(NebulaTunnelInfo *info) {
    if (info->bException) {

        P2PTunnelAgent_Disconnect(info->nSID);

        P2PTunnelAgent_StopPortMapping(info->nHttpIndex);
#if ENABLE_TUNNEL_SERVICE
        P2PTunnelAgent_StopServiceMapping(info->nHttpsIndex);
#endif
        P2PTunnelAgent_StopPortMapping(info->nTelnetIndex);
        P2PTunnelAgent_StopPortMapping(info->nSshIndex);

        IOTC_Session_Close(info->nSID); // if AV module working on this session, AV will be disconnected.

        info->nSID = -1;
        info->bException = false;

        return true;
    }
    return false;
}

static void WINAPI tunnelStatusHandle(int nErrorCode, int nSID, void *pArg) {
    NebulaTunnelInfo *info = (NebulaTunnelInfo *)pArg;
    if (nErrorCode == TUNNEL_ER_DISCONNECTED) {
        printf("SID[%d] TUNNEL_ER_DISCONNECTED\n", nSID);
        if (info->nSID == nSID) {
            info->bException = true;
        }
    }
}

static int tunnelAgentConnect(NebulaTunnelInfo *info) {
    int ret = P2PTunnelSetConnectionOption(TUNNEL_CONNECT_MANUAL);
    if (ret != IOTC_ER_NoERROR) {
        printf("P2PTunnelSetConnectionOption() ret[%d]\n", ret);
        return -1;
    }

    ret = P2PTunnelAgentInitialize2(1, 0);
    if (ret != TUNNEL_ER_NoERROR) {
        printf("P2PTunnelAgentInitialize2() ret[%d]\n", ret);
        return -1;
    }

    P2PTunnelAgent_GetStatus(tunnelStatusHandle, info);

    ret = P2PTunnelAgent_Attach_Connect_Ex(info->nSID, 1, NULL, NULL);
    if (ret < 0) {
        printf("tunnel attach connect error[%d]\n", ret);
        return -1;
    } else {
        printf("tunnel attach connect success\n");
    }

    ret = P2PTunnel_SetBufSize(info->nSID, 5120000);
    if (ret != TUNNEL_ER_NoERROR) {
        printf("P2PTunnel_SetBufSize() sid[%d] error[%d]\n", info->nSID, ret);
    }

    info->nHttpIndex = P2PTunnelAgent_PortMapping(info->nSID, HTTP_MAPPING_LOCAL_PORT, HTTP_MAPPING_REMOTE_PORT);
    if (info->nHttpIndex < 0) {
        printf("P2PTunnelAgent_PortMapping() HTTP sid[%d] error[%d]!\n", info->nSID, info->nHttpIndex);
    } else {
        printf("sid[%d] HTTP mapping success, LOCAL[%d]<->REMOTE[%d]\n", info->nSID, HTTP_MAPPING_LOCAL_PORT, HTTP_MAPPING_REMOTE_PORT);
    }

#if ENABLE_TUNNEL_SERVICE
    info->nHttpsIndex = P2PTunnelAgent_ServiceMapping(info->nSID, HTTPS_MAPPING_LOCAL_PORT, "https");
    if (info->nHttpsIndex < 0) {
        printf("P2PTunnelAgent_ServiceMapping() HTTPS sid[%d] error[%d]!\n", info->nSID, info->nHttpsIndex);
    } else {
        printf("sid[%d] HTTPS mapping success, LOCAL[%d]<->REMOTE[https]\n", info->nSID, HTTPS_MAPPING_LOCAL_PORT);
    }
#endif

    info->nTelnetIndex = P2PTunnelAgent_PortMapping(info->nSID, TELNET_MAPPING_LOCAL_PORT, TELNET_MAPPING_REMOTE_PORT);
    if (info->nTelnetIndex < 0) {
        printf("P2PTunnelAgent_PortMapping() TELENT sid[%d] error[%d]!\n", info->nSID, info->nTelnetIndex);
    } else {
        printf("sid[%d] TELENT mapping success, LOCAL[%d]<->REMOTE[%d]\n", info->nSID, TELNET_MAPPING_LOCAL_PORT, TELNET_MAPPING_REMOTE_PORT);
    }

    info->nSshIndex = P2PTunnelAgent_PortMapping(info->nSID, SSH_MAPPING_LOCAL_PORT, SSH_MAPPING_REMOTE_PORT);
    if (info->nSshIndex < 0) {
        printf("P2PTunnelAgent_PortMapping() SSH sid[%d] error[%d]!\n", info->nSID, info->nSshIndex);
    } else {
        printf("sid[%d] SSH mapping success, LOCAL[%d]<->REMOTE[%d]\n", info->nSID, SSH_MAPPING_LOCAL_PORT, SSH_MAPPING_REMOTE_PORT);
    }

    info->nRetryConnectFailedCnt = 0;
    info->bAgentSuccess = true;

    return 0;
}

//########################################################
//# Main
//########################################################
int main(int argc, char *argv[]) {
    int ret = 0;
    char *udid = NULL;
    char *am_token = NULL;
    static char pin_code[MAX_PIN_CODE_LENGTH + 1] = {0};
    char *license_key = NULL;
    LogAttr log_attr = {0};
    char log_path[] = "./client_log.txt";
    NebulaClientCtx *client_ctx = NULL;
    NebulaTunnelInfo *nebula_tunnel_info = calloc(1, sizeof(NebulaTunnelInfo));

    nebula_tunnel_info->pClientCtx = NULL;
    nebula_tunnel_info->bAgentSuccess = false;
    nebula_tunnel_info->nHttpIndex = -1;
    nebula_tunnel_info->nHttpsIndex = -1;
    nebula_tunnel_info->nTelnetIndex = -1;
    nebula_tunnel_info->nSshIndex = -1;
    nebula_tunnel_info->nRetryConnectFailedCnt = 0;
    nebula_tunnel_info->bException = false;

    if (argc != 4) {
        printf("%s <LICENSE_KEY> <UDID> <AM Token>\n", argv[0]);
        return -1;
    }

    license_key = argv[1];
    udid = argv[2];
    am_token = argv[3];

    if (strlen(udid) != MAX_PUBLIC_UDID_LENGTH) {
        printf("invalid UDID\n");
        free(nebula_tunnel_info);
        return -1;
    }

    //===== get disposable params for Nebula connect =====
    ret = loadDisposableParams(pin_code, NULL, NULL);
    if (ret < 0) {
        printf("loadDisposableParams()=%d\n", ret);
        free(nebula_tunnel_info);
        return -1;
    }

    //===== set license key for enable TUTK SDK =====
    ret = TUTK_SDK_Set_License_Key(license_key);
    printf("TUTK_SDK_Set_License_Key() ret[%d]\n", ret);
    if (ret < 0) {
        free(nebula_tunnel_info);
        return -1;
    }

    //===== create Nebula client =====
    client_ctx = nebulaclientCreate(udid, pin_code);
    if (client_ctx == NULL) {
        free(nebula_tunnel_info);
        return -1;
    }
    nebula_tunnel_info->pClientCtx = client_ctx;

    //===== create IOTC client by Nebula =====
    ret = IOTC_Initialize2(0);
    if (ret != IOTC_ER_NoERROR) {
        printf("IOTC_Initialize2() ret[%d]\n", ret);
        return -1;
    }

    ret = iotcClientCreateByNebula(client_ctx, am_token, nebula_tunnel_info, &gIOTCConnectAbort);
    if (ret < 0) {
        IOTC_DeInitialize();
        Nebula_Client_Delete(client_ctx);
        Nebula_DeInitialize();
        free(nebula_tunnel_info);
        return -1;
    }

    //==== create P2PTunnel agent by Nebula =====
    log_attr.path = log_path;
    ret = P2PTunnel_Set_Log_Attr(log_attr);
    printf("P2PTunnel_Set_Log_Attr() ret[%d]\n", ret);

    ret = tunnelAgentConnect(nebula_tunnel_info);
    if (ret <0) {
        P2PTunnelAgentDeInitialize();
        IOTC_Session_Close(nebula_tunnel_info->nSID); // if AV module working on this session, AV will be disconnected.
        IOTC_DeInitialize();
        Nebula_Client_Delete(client_ctx);
        Nebula_DeInitialize();
        free(nebula_tunnel_info);
        return -1;
    }

    while (1) {
        if (nebula_tunnel_info->nSID >= 0) {
            if (nebula_tunnel_info->bAgentSuccess) {
                int access_time = P2PTunnel_LastIOTime(nebula_tunnel_info->nSID);
                if (access_time < 0) {
                    printf("P2PTunnel_LastIOTime() sid[%d] error[%d]\n", nebula_tunnel_info->nSID, access_time);
                } else {
                    printf("sid[%d]:%u, \n", nebula_tunnel_info->nSID, access_time);
                }
            } else {
                printf("Reconnect to sid[%d]\n", nebula_tunnel_info->nSID);
                tunnelAgentConnect(nebula_tunnel_info);

                if (++nebula_tunnel_info->nRetryConnectFailedCnt > 3) {
                    printf("Retry connection timeout sid[%d]\n", nebula_tunnel_info->nSID);
                    nebula_tunnel_info->bException = true;
                }
            }
        }
        Sleep(1000);
        if (tunnelExceptionCheck(nebula_tunnel_info) == true) {
            break;
        }
    }

    P2PTunnelAgent_Disconnect(nebula_tunnel_info->nSID);
    P2PTunnelAgentDeInitialize();

    IOTC_Session_Close(nebula_tunnel_info->nSID);
    IOTC_DeInitialize();

    Nebula_Client_Delete(client_ctx);
    Nebula_DeInitialize();
    free(nebula_tunnel_info);
    return 0;
}
